#include <string>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <windows.h>
#include "tclap/CmdLine.h"
#include "modbus.h"
#include "modbus-rtu.h"
#include "BZP/KIWIDatabaseIDs.h"
#include "ByteOrderingConverter.h"
#include "ModbusFunction65.h"


/*========================================   ==========================================*/

enum class ActionType
{
	AT_UNDEFINED,
	AT_UPLOAD_FILE,
	AT_DOWNLOAD_FILE,
	AT_DOWNLOAD_OSCILLOGRAM,
	AT_SHOW_OSCILLOGRAM_LIST,
	AT_READ_EVENT_LOG,
	AT_READ_MESSAGE_LOG,
	AT_READ_SETTINGS_LOG
};

/*========================================  ===========================================*/

static ActionType s_actionType = ActionType::AT_UNDEFINED;
static std::string s_comPortName = "\\\\.\\";
static int s_baudRate = 115200;
static int s_serverAddress = 1;
static std::string s_remoteFilename;
static std::string s_localFilename;
static mbf65_CompressionType s_compressionType = MBF65_CT_ZLIB_DEFLATE_BEST_SPEED;
static int s_oscillogramNumber = 1;
static int s_firstRecordNumber = 1;
static int s_numberOfRecords = 1;

/*=====================================   =======================================*/

static int parseCommandLine(int argc, char** argv);

static int run(modbus_t* pModbus, const char** ppErrorMessage);

static int showOscillogramList(modbus_t* pModbus, uint8_t address);

static int readEventLog(modbus_t* pModbus, uint8_t address, int firstRecordNumber, 
	int numberOfRecords);

static int readMessageLog(modbus_t* pModbus, uint8_t address, int firstRecordNumber,
	int numberOfRecords);

static int readSettingsLog(modbus_t* pModbus, uint8_t address, int firstRecordNumber,
	int numberOfRecords);

static int printTimestampWithMilliseconds(char* pBuffer, uint64_t timestamp);

static int readRegisters(modbus_t* pModbus, uint16_t address, uint16_t numberOfRegisters);

static void printEventLogDescriptor(void* pDescriptor);

static void* printIDArray(uint16_t* pIDArray, const char* pTitle);

static void printEventLogRecords(const void* pDescriptor, void* pRecords, uint32_t numberOfRecords);

static const void* printBinarySignals(const void* pBitArray, const uint16_t* pIDArray,
	uint16_t numberOfSignals, const char* pTitle);

static const void* printAnalogOutputSignals(const float* pValues,
	const mbf65_EventLogAOSDescriptorExp* pDescriptors, uint16_t numberOfSignals, 
	const char* pTitle);

static const void* printAccumulators(const uint32_t* pValues, const uint16_t* pIDArray,
	uint16_t numberOfSignals, const char* pTitle);

/*==========================================  ============================================*/

int main(int argc, char** argv)
{
	int rc = 0;
	const char* pErrorMessage = NULL;

	rc = parseCommandLine(argc, argv);

	if (rc == 0)
	{
		char parity = 'N';
		int data_bits = 8;
		int stop_bits = 1;
		modbus_t* pModbus;

		fprintf(stdout,
			"Action = %u, Remote file=\"%s\", Local file=\"%s\", Port=%s, address=%u, baud rate=%u, compression=%u\n",
			s_actionType, s_remoteFilename.c_str(), s_localFilename.c_str(), s_comPortName.c_str(),
			s_serverAddress, s_baudRate, s_compressionType);

		pModbus = modbus_new_rtu(s_comPortName.c_str(), s_baudRate, parity, data_bits, stop_bits);
		if (pModbus != NULL)
		{
			struct timeval timeout;

			modbus_set_debug(pModbus, FALSE);

			timeout.tv_sec = 1;
			timeout.tv_usec = 0;
			modbus_set_response_timeout(pModbus, &timeout);

			timeout.tv_sec = 0;
			timeout.tv_usec = 50000;
			modbus_set_byte_timeout(pModbus, &timeout);

			rc = modbus_set_slave(pModbus, s_serverAddress);
			pErrorMessage = "modbus_set_slave() failed: %s\n";

			if (rc == 0)
			{
				rc = modbus_connect(pModbus);
				pErrorMessage = "Connection failed: %s\n";

				if (rc == 0)
				{
					rc = run(pModbus, &pErrorMessage);
					modbus_close(pModbus);
				} /* if (modbus_connect(pModbus) == 0) */
			} /* if (rc == 0) */

			modbus_free(pModbus);
		} /* if (pModbus != NULL) */
		else
		{
			rc = -1;
			pErrorMessage = "modbus_new_rtu() failed: %s\n";
		}
	}

	if (pErrorMessage != NULL && rc != 0)
	{
		fprintf(stderr, pErrorMessage, modbus_strerror(errno));
	}

	return rc;
} /* main() */
/*-----------------------------------------------------------------------------------------------*/
int parseCommandLine(int argc, char** argv)
{
	int rc = 0;

	try
	{
		TCLAP::CmdLine commandLine("", ' ', "0.1", true);

		TCLAP::SwitchArg uploadArg("", "upload", "Upload file to device", false);
		TCLAP::SwitchArg downloadArg("", "download", "Download file to device", false);
		TCLAP::SwitchArg getoscArg("", "getosc", "Download oscillogram", false);
		TCLAP::SwitchArg osclistArg("", "osclist", "Show oscillogram list", false);
		TCLAP::SwitchArg eventLogArg("", "eventlog", "Show records from event log", false);
		TCLAP::SwitchArg messageLogArg("", "msglog", "Show records from message log", false);
		TCLAP::SwitchArg settingsLogArg("", "setlog", "Show records from settings log", false);
		std::vector<TCLAP::Arg*> actionArgs;

		actionArgs.push_back(&uploadArg);
		actionArgs.push_back(&downloadArg);
		actionArgs.push_back(&getoscArg);
		actionArgs.push_back(&osclistArg);
		actionArgs.push_back(&eventLogArg);
		actionArgs.push_back(&messageLogArg);
		actionArgs.push_back(&settingsLogArg);

		TCLAP::ValueArg<std::string> remoteFileArg("", "remote", "Remote file name", false, "", "remote_file");
		TCLAP::ValueArg<std::string> localFileArg("", "local", "Local file name", false, "", "local_file");
		TCLAP::ValueArg<std::string> comPortArg("p", "port", "COM port", true, "", "com_port");
		TCLAP::ValueArg<int> compressionLevelArg("c", "compression", "Compression level", false, 3, "compression_level");
		TCLAP::ValueArg<int> modbusAddressArg("a", "address", "Modbus address", false, 1, "modbus_address");
		TCLAP::ValueArg<int> baudRateArg("b", "baud", "Baud rate", false, 115200, "baud_rate");
		TCLAP::ValueArg<int> oscillogramNumberArg("", "oscnumber", "Oscillogram number", false, 0, "oscillogram_number");
		TCLAP::ValueArg<int> firstRecordNumberArg("", "firstrec", "First record number", false, 0, "first_record_number");
		TCLAP::ValueArg<int> numberOfRecordsArg("", "records", "Number of records", false, 0, "number_of_records");

		commandLine.xorAdd(actionArgs);
		commandLine.add(remoteFileArg);
		commandLine.add(localFileArg);
		commandLine.add(oscillogramNumberArg);
		commandLine.add(firstRecordNumberArg);
		commandLine.add(numberOfRecordsArg);
		commandLine.add(comPortArg);
		commandLine.add(compressionLevelArg);
		commandLine.add(modbusAddressArg);
		commandLine.add(baudRateArg);

		commandLine.parse(argc, argv);

		s_remoteFilename = remoteFileArg.getValue();
		s_localFilename = localFileArg.getValue();
		s_oscillogramNumber = oscillogramNumberArg.getValue();
		s_firstRecordNumber = firstRecordNumberArg.getValue();
		s_numberOfRecords = numberOfRecordsArg.getValue();
		s_compressionType = (mbf65_CompressionType)compressionLevelArg.getValue();

		s_comPortName += comPortArg.getValue();
		s_baudRate = baudRateArg.getValue();
		s_serverAddress = modbusAddressArg.getValue();

		if (uploadArg.getValue())
		{
			s_actionType = ActionType::AT_UPLOAD_FILE;
		}
		else if (downloadArg.getValue())
		{
			s_actionType = ActionType::AT_DOWNLOAD_FILE;
		}
		else if (getoscArg.getValue())
		{
			s_actionType = ActionType::AT_DOWNLOAD_OSCILLOGRAM;
		}
		else if (osclistArg.getValue())
		{
			s_actionType = ActionType::AT_SHOW_OSCILLOGRAM_LIST;
		}
		else if (eventLogArg.getValue())
		{
			s_actionType = ActionType::AT_READ_EVENT_LOG;
		}
		else if (messageLogArg.getValue())
		{
			s_actionType = ActionType::AT_READ_MESSAGE_LOG;
		}
		else if (settingsLogArg.getValue())
		{
			s_actionType = ActionType::AT_READ_SETTINGS_LOG;
		}
	}
	catch (TCLAP::ArgException& e)
	{
		rc = -1;
		fprintf(stdout, "TCLAP exception: %s\n", e.what());
	}

	return rc;
} /* parseCommandLine() */
/*-----------------------------------------------------------------------------------------------*/
int run(modbus_t* pModbus, const char** ppErrorMessage)
{
	int rc = -1;
	float operationTime;
	uint32_t originalFileSize;
	uint32_t compressedFileSize;
	clock_t t;

	switch (s_actionType)
	{
	case ActionType::AT_UPLOAD_FILE:
		t = clock();

		rc = mbf65_uploadFile(pModbus, s_serverAddress, s_remoteFilename.c_str(),
			s_localFilename.c_str(), &originalFileSize);

		t = clock() - t;
		operationTime = (float)t / (float)CLOCKS_PER_SEC;
		if (rc == 0)
		{
			fprintf(stdout,
				"File \"%s\" uploaded in %.3f s, size=%u, speed=%u bytes/s\n",
				s_remoteFilename.c_str(), operationTime, originalFileSize,
				(uint32_t)(originalFileSize / operationTime));
		}
		*ppErrorMessage = "uploadFile() failed: %s\n";
		break;
	case ActionType::AT_DOWNLOAD_FILE:
		if (s_compressionType == 0)
		{
			t = clock();

			rc = mbf65_downloadFile(pModbus, s_serverAddress, s_remoteFilename.c_str(),
				s_localFilename.c_str(), &originalFileSize);

			t = clock() - t;
			operationTime = (float)t / (float)CLOCKS_PER_SEC;
			if (rc == 0)
			{
				fprintf(stdout,
					"File \"%s\" downloaded in %.3f s, size=%u, speed=%u bytes/s\n",
					s_remoteFilename.c_str(), operationTime, originalFileSize,
					(uint32_t)(originalFileSize / operationTime));
			}
			*ppErrorMessage = "downloadFile() failed: %s\n";
		}
		else
		{
			t = clock();

			rc = mbf65_downloadFileCompressed(pModbus, s_serverAddress,
				s_remoteFilename.c_str(), s_localFilename.c_str(), s_compressionType,
				&originalFileSize, &compressedFileSize);

			t = clock() - t;
			operationTime = (float)t / (float)CLOCKS_PER_SEC;
			if (rc == 0)
			{
				fprintf(stdout,
					"File \"%s\" downloaded in %.3f s, compression type=%u, "
					"original size=%u, compressed size=%u, compression ratio=%.1f, "
					"original speed=%u bytes/s, compressed speed=%u bytes/s\n",
					s_remoteFilename.c_str(), operationTime, s_compressionType,
					originalFileSize, compressedFileSize,
					(float)originalFileSize / (float)compressedFileSize,
					(uint32_t)(originalFileSize / operationTime),
					(uint32_t)(compressedFileSize / operationTime));
			}
			*ppErrorMessage = "downloadFileCompressed() failed: %s\n";
		}
		break;
	case ActionType::AT_DOWNLOAD_OSCILLOGRAM:
		if (s_compressionType == 0)
		{
			t = clock();

			rc = mbf65_downloadOscillogram(pModbus, s_serverAddress, s_oscillogramNumber,
				s_localFilename.c_str(), &originalFileSize);

			t = clock() - t;
			operationTime = (float)t / (float)CLOCKS_PER_SEC;
			if (rc == 0)
			{
				fprintf(stdout,
					"Oscillogram %u downloaded in %.3f s, size=%u, speed=%u bytes/s\n",
					s_oscillogramNumber, operationTime, originalFileSize,
					(uint32_t)(originalFileSize / operationTime));
			}
			*ppErrorMessage = "mbf65_downloadOscillogram() failed: %s\n";
		}
		else
		{
			t = clock();

			rc = mbf65_downloadOscillogramCompressed(pModbus, s_serverAddress,
				s_oscillogramNumber, s_localFilename.c_str(), s_compressionType,
				&originalFileSize, &compressedFileSize);

			t = clock() - t;
			operationTime = (float)t / (float)CLOCKS_PER_SEC;
			if (rc == 0)
			{
				fprintf(stdout,
					"Oscillogram %u downloaded in %.3f s, compression type=%u, "
					"original size=%u, compressed size=%u, compression ratio=%.1f, "
					"original speed=%u bytes/s, compressed speed=%u bytes/s\n",
					s_oscillogramNumber, operationTime, s_compressionType,
					originalFileSize, compressedFileSize,
					(float)originalFileSize / (float)compressedFileSize,
					(uint32_t)(originalFileSize / operationTime),
					(uint32_t)(compressedFileSize / operationTime));
			}
			*ppErrorMessage = "mbf65_downloadOscillogramCompressed() failed: %s\n";
		}
		break;
	case ActionType::AT_SHOW_OSCILLOGRAM_LIST:
		rc = showOscillogramList(pModbus, s_serverAddress);
		break;
	case ActionType::AT_READ_EVENT_LOG:
		rc = readEventLog(pModbus, s_serverAddress, s_firstRecordNumber, s_numberOfRecords);
		break;
	case ActionType::AT_READ_MESSAGE_LOG:
		rc = readMessageLog(pModbus, s_serverAddress, s_firstRecordNumber, s_numberOfRecords);
		break;
	case ActionType::AT_READ_SETTINGS_LOG:
		rc = readSettingsLog(pModbus, s_serverAddress, s_firstRecordNumber, s_numberOfRecords);
		break;
	}

	return rc;
} /* run() */
/*-----------------------------------------------------------------------------------------------*/
int showOscillogramList(modbus_t* pModbus, uint8_t address)
{
	int rc;
	uint16_t numberOfOscillograms;

	rc = mbf65_readNumberOfOscillograms(pModbus, address, &numberOfOscillograms);
	if (rc == 0)
	{
		uint16_t numberOfDescriptorsRead = 0;
		mbf65_OscillogramInfoExp* pDescriptors = 
			(mbf65_OscillogramInfoExp*)malloc(numberOfOscillograms * sizeof(mbf65_OscillogramInfoExp));

		rc = mbf65_readOscillogramDescriptors(pModbus, address, 0, numberOfOscillograms, 
			pDescriptors, &numberOfDescriptorsRead);

		if (rc == 0)
		{
			for (uint16_t i = 0; i < numberOfDescriptorsRead; i++)
			{
				char pTimestamp[40];

				printTimestampWithMilliseconds(pTimestamp, pDescriptors[i].timestamp);
				fprintf(stdout, "%u: timestamp=%s, duration=%u ms, size=%u byte, signal type=%u, signal ID=%u\n", 
					i + 1, pTimestamp, pDescriptors[i].duration, pDescriptors[i].fileSize, 
					pDescriptors[i].signalType, pDescriptors[i].signalID);
			}
		}

		free(pDescriptors);
	}

	return rc;
} /* showOscillogramList() */
/*-----------------------------------------------------------------------------------------------*/
int readEventLog(modbus_t* pModbus, uint8_t address, int firstRecordNumber,
	int numberOfRecords)
{
	int rc = 0;
	mbf65_LogInfoExp logInfo;

	rc = mbf65_readLogInfo(pModbus, address, MBF65_LT_EVENT_LOG, &logInfo);

	if (rc == 0)
	{
		const uint32_t descriptorBufferSize = 30000;
		void* pDescriptorBuffer = malloc(descriptorBufferSize);
		uint32_t numberOfBytesRead = 0;
		char pTimestamp[40];

		printTimestampWithMilliseconds(pTimestamp, logInfo.lastRecordTimestamp);
		fprintf(stdout, "logVersion=%u, softwareVersion=%u, recordSize=%u, maxNumberOfRecords=%u, "
			"currentNumberOfRecords=%u, lastRecordID=%llu, lastRecordTimestamp=%s\n",
			logInfo.logVersion, logInfo.softwareVersion, logInfo.recordSize, 
			logInfo.maxNumberOfRecords, logInfo.currentNumberOfRecords, logInfo.lastRecordID, 
			pTimestamp);

		fprintf(stdout, "\nReading event log descriptor:\n");

		rc = mbf65_readDescriptorOfEventLogRecord(pModbus, address, pDescriptorBuffer, 
			descriptorBufferSize, &numberOfBytesRead);

		if (rc == 0)
		{
			uint32_t numberOfRecordsRead = 0;
			const uint32_t recordBufferSize = numberOfRecords * logInfo.recordSize;
			void* pRecordBuffer = malloc(recordBufferSize);

			printEventLogDescriptor(pDescriptorBuffer);

			fprintf(stdout, "\nReading event log records:\n");

			rc = mbf65_readFullLogRecords(pModbus, address, MBF65_LT_EVENT_LOG,
				logInfo.recordSize, firstRecordNumber, numberOfRecords, pRecordBuffer, 
				recordBufferSize, &numberOfRecordsRead);
			if (rc == 0)
			{
				printEventLogRecords(pDescriptorBuffer, pRecordBuffer, numberOfRecordsRead);
			}

			free(pRecordBuffer);
		}

		free(pDescriptorBuffer);
	}

	return rc;
} /* readEventLog() */
/*-----------------------------------------------------------------------------------------------*/
int readMessageLog(modbus_t* pModbus, uint8_t address, int firstRecordNumber,
	int numberOfRecords)
{
	int rc = 0;
	mbf65_LogInfoExp logInfo;

	rc = mbf65_readLogInfo(pModbus, address, MBF65_LT_MESSAGE_LOG, &logInfo);

	if (rc == 0)
	{
		char pTimestamp[40];
		uint32_t numberOfRecordsRead = 0;
		const uint32_t recordBufferSize = numberOfRecords * logInfo.recordSize;
		void* pRecordBuffer = malloc(recordBufferSize);

		printTimestampWithMilliseconds(pTimestamp, logInfo.lastRecordTimestamp);
		fprintf(stdout, "logVersion=%u, softwareVersion=%u, recordSize=%u, maxNumberOfRecords=%u, "
			"currentNumberOfRecords=%u, lastRecordID=%llu, lastRecordTimestamp=%s\n",
			logInfo.logVersion, logInfo.softwareVersion, logInfo.recordSize,
			logInfo.maxNumberOfRecords, logInfo.currentNumberOfRecords, logInfo.lastRecordID,
			pTimestamp);

		fprintf(stdout, "\nReading message log records:\n");

		rc = mbf65_readFullLogRecords(pModbus, address, MBF65_LT_MESSAGE_LOG,
			logInfo.recordSize, firstRecordNumber, numberOfRecords, pRecordBuffer,
			recordBufferSize, &numberOfRecordsRead);

		if (rc == 0)
		{
			msglog_MessageLogRecordBuffer* pRecord =
				(msglog_MessageLogRecordBuffer*)pRecordBuffer;
			for (uint32_t i = 0; i < numberOfRecordsRead; i++)
			{
				boc_convert64BitValueInSitu(&pRecord->heading.timestamp);
				boc_convert64BitValueInSitu(&pRecord->heading.recordNo);

				printTimestampWithMilliseconds(pTimestamp, pRecord->heading.timestamp);
				fprintf(stdout, "%u: recNo=%llu, timestamp=%s, message=\"%s\"\n", i + 1, 
					pRecord->heading.recordNo, pTimestamp, pRecord->pMessage);

				pRecord = (msglog_MessageLogRecordBuffer*)((uint8_t*)pRecord + logInfo.recordSize);
			}
		}

		free(pRecordBuffer);
	}

	return rc;
} /* readMessageLog() */
/*-----------------------------------------------------------------------------------------------*/
int readSettingsLog(modbus_t* pModbus, uint8_t address, int firstRecordNumber,
	int numberOfRecords)
{
	int rc = 0;
	mbf65_LogInfoExp logInfo;

	rc = mbf65_readLogInfo(pModbus, address, MBF65_LT_SETTINGS_LOG, &logInfo);

	if (rc == 0)
	{
		char pTimestamp[40];
		uint32_t numberOfRecordsRead = 0;
		const uint32_t recordBufferSize = numberOfRecords * logInfo.recordSize;
		void* pRecordBuffer = malloc(recordBufferSize);

		printTimestampWithMilliseconds(pTimestamp, logInfo.lastRecordTimestamp);
		fprintf(stdout, "logVersion=%u, softwareVersion=%u, recordSize=%u, maxNumberOfRecords=%u, "
			"currentNumberOfRecords=%u, lastRecordID=%llu, lastRecordTimestamp=%s\n",
			logInfo.logVersion, logInfo.softwareVersion, logInfo.recordSize,
			logInfo.maxNumberOfRecords, logInfo.currentNumberOfRecords, logInfo.lastRecordID,
			pTimestamp);

		fprintf(stdout, "\nReading settings log records:\n");

		rc = mbf65_readFullLogRecords(pModbus, address, MBF65_LT_SETTINGS_LOG,
			logInfo.recordSize, firstRecordNumber, numberOfRecords, pRecordBuffer,
			recordBufferSize, &numberOfRecordsRead);

		if (rc == 0)
		{
			void* pRecord = pRecordBuffer;
			for (uint32_t i = 0; i < numberOfRecordsRead; i++)
			{
				if (logInfo.logVersion == 1 || logInfo.logVersion == 2)
				{
					setlog_Record_ver_1_and_2* pRecordV1_2 = (setlog_Record_ver_1_and_2*)pRecord;

					boc_convert64BitValueInSitu(&pRecordV1_2->heading.timestamp);
					boc_convert64BitValueInSitu(&pRecordV1_2->heading.recordNo);
					printTimestampWithMilliseconds(pTimestamp, pRecordV1_2->heading.timestamp);
					fprintf(stdout, "%u: recNo=%llu, timestamp=%s, settingID=%u, sourceID=%u, oldValue=%d, newValue=%d\n",
						i + 1, pRecordV1_2->heading.recordNo, pTimestamp,
						*boc_convert16BitValueInSitu(&pRecordV1_2->settingID), pRecordV1_2->sourceID,
						(int32_t)*boc_convert32BitValueInSitu(&pRecordV1_2->oldValue),
						(int32_t)*boc_convert32BitValueInSitu(&pRecordV1_2->newValue));
				}
				else
				{
					setlog_Record_ver_3* pRecordV3 = (setlog_Record_ver_3*)pRecord;

					boc_convert64BitValueInSitu(&pRecordV3->heading.timestamp);
					boc_convert64BitValueInSitu(&pRecordV3->heading.recordNo);
					printTimestampWithMilliseconds(pTimestamp, pRecordV3->heading.timestamp);
					fprintf(stdout, "%u: recNo=%llu, timestamp=%s, settingID=%u, sourceID=%u, program=%u, oldValue=%d, newValue=%d\n",
						i + 1, pRecordV3->heading.recordNo, pTimestamp,
						*boc_convert16BitValueInSitu(&pRecordV3->settingID), pRecordV3->sourceID, 
						pRecordV3->programNo, 
						(int32_t)*boc_convert32BitValueInSitu(&pRecordV3->oldValue),
						(int32_t)*boc_convert32BitValueInSitu(&pRecordV3->newValue));
				}

				pRecord = (uint8_t*)pRecord + logInfo.recordSize;
			}
		}

		free(pRecordBuffer);
	}

	return rc;
} /* readSettingsLog() */
/*-----------------------------------------------------------------------------------------------*/
int printTimestampWithMilliseconds(char* pBuffer, uint64_t timestamp)
{
	int n = 0;
	int milliseconds = (int)(timestamp % 1000ULL);
	time_t t = (time_t)(timestamp / 1000ULL);
	struct tm* pTm = gmtime(&t);

	if (pTm == NULL)
	{
		n = sprintf(pBuffer, "%llu.%03u", (uint64_t)t, milliseconds);
	}
	else
	{
		n = sprintf(pBuffer, "%04u-%02u-%02u %02u:%02u:%02u.%03u",
			pTm->tm_year + 1900, pTm->tm_mon + 1, pTm->tm_mday, pTm->tm_hour, pTm->tm_min,
			pTm->tm_sec, milliseconds);
	}

	return n;
} /* printTimestampWithMilliseconds() */
/*-----------------------------------------------------------------------------------------------*/
int readRegisters(modbus_t* pModbus, uint16_t address, uint16_t numberOfRegisters)
{
	int rc = -1;
	uint16_t* pRegisters;

	pRegisters = (uint16_t*)malloc(numberOfRegisters * sizeof(uint16_t));
	if (pRegisters != NULL)
	{
		rc = modbus_read_registers(pModbus, address, numberOfRegisters, pRegisters);

		if (rc > 0)
		{
			for (int i = 0; i < rc; i++)
			{
				fprintf(stdout, "[0x%04X] = 0x%04X\n", address + i, pRegisters[i]);
			}
		}

		free(pRegisters);
	}

	return rc < 0 ? rc : 0;
} /* readRegisters() */
/*-----------------------------------------------------------------------------------------------*/
void printEventLogDescriptor(void* pDescriptor)
{
	/* . */
	uint16_t numberOfEventTypes = *(uint16_t*)pDescriptor;
	pDescriptor = (uint8_t*)pDescriptor + sizeof(numberOfEventTypes);
	fprintf(stdout, "Events:\nnumberOfEventTypes=%u\n", numberOfEventTypes);
	for (uint16_t i = 0; i < numberOfEventTypes; i++)
	{
		fprintf(stdout, "id=%u, name=\"%s\"\n", 
			((mbf65_EventLogEventDescriptorExp*)pDescriptor)->id,
			((mbf65_EventLogEventDescriptorExp*)pDescriptor)->name);
		pDescriptor = (uint8_t*)pDescriptor + sizeof(mbf65_EventLogEventDescriptorExp);
	}

	/*  . */
	pDescriptor = printIDArray((uint16_t*)pDescriptor, "Binary Inputs:");

	/*  . */
	pDescriptor = printIDArray((uint16_t*)pDescriptor, "Binary Outputs:");

	/*   . */
	pDescriptor = printIDArray((uint16_t*)pDescriptor, "Binary Input Signals:");

	/*   . */
	pDescriptor = printIDArray((uint16_t*)pDescriptor, "Binary Output Signals:");

	/*   . */
	uint16_t numberOfAOS = *(uint16_t*)pDescriptor;
	pDescriptor = (uint8_t*)pDescriptor + sizeof(numberOfAOS);
	fprintf(stdout, "\nAnalog Output Signals:\n");
	fprintf(stdout, "numberOfIDs=%u\n", numberOfAOS);
	for (uint16_t i = 0; i < numberOfAOS; i++)
	{
		fprintf(stdout, "ID%u=%u, units=%u\n", i + 1, 
			((mbf65_EventLogAOSDescriptorExp*)pDescriptor)->id,
			((mbf65_EventLogAOSDescriptorExp*)pDescriptor)->units);
		pDescriptor = (uint8_t*)pDescriptor + sizeof(mbf65_EventLogAOSDescriptorExp);
	}

	/* . */
	pDescriptor = printIDArray((uint16_t*)pDescriptor, "Accumulators:");
} /* convertEventLogDescriptorToHostEndianness() */
/*-----------------------------------------------------------------------------------------------*/
void* printIDArray(uint16_t* pIDArray, const char* pTitle)
{
	uint16_t numberOfIDs = *pIDArray++;

	fprintf(stdout, "\n%s\n", pTitle);
	fprintf(stdout, "numberOfIDs=%u\n", numberOfIDs);

	for (uint16_t i = 0; i < numberOfIDs; i++)
	{
		fprintf(stdout, "ID%u=%u\n", i + 1,  *pIDArray++);
	}

	return pIDArray;
} /* printIDArray() */
/*-----------------------------------------------------------------------------------------------*/
void printEventLogRecords(const void* pDescriptor, void* pRecords, uint32_t numberOfRecords)
{
	mbf65_EventLogRecordBuffer* pRecord = (mbf65_EventLogRecordBuffer*)pRecords;
	uint16_t numberOfEvents = *(uint16_t*)pDescriptor;
	const mbf65_EventLogEventDescriptorExp* pEventDescriptor = 
		(const mbf65_EventLogEventDescriptorExp*)((const uint8_t*)pDescriptor + sizeof(uint16_t));
	
	for (uint32_t i = 0; i < numberOfRecords; i++)
	{
		char pTimestamp[40];
		void* pData = pRecord->pData;

		boc_convert64BitValueInSitu(&pRecord->heading.timestamp);
		boc_convert64BitValueInSitu(&pRecord->heading.recordNo);

		printTimestampWithMilliseconds(pTimestamp, pRecord->heading.timestamp);
		fprintf(stdout, "%u: recNo=%llu, timestamp=%s\n", i + 1, pRecord->heading.recordNo, 
			pTimestamp);

		/*   . */
		fprintf(stdout, "Events:\n");
		for (uint16_t eventIndex = 0; eventIndex < numberOfEvents; eventIndex++)
		{
			bool isEventSet = mbf65_getBinaryValueFromBitArray(pData, eventIndex) != 0;
			if (isEventSet)
			{
				fprintf(stdout, "i=%u, id=%u, name=\"%s\"\n", eventIndex, pEventDescriptor->id, 
					pEventDescriptor->name);
			}

			pEventDescriptor++;
		}
		pData = (uint8_t*)pData + numberOfEvents / 8 + 1;

		const uint8_t* pIDs = (const uint8_t*)pEventDescriptor;

		/*    . */
		pData = (void*)printBinarySignals(pData, (uint16_t*)(pIDs + sizeof(uint16_t)), 
			*(uint16_t*)pIDs, "Binary Inputs");
		pIDs += *(uint16_t*)pIDs * sizeof(uint16_t) + sizeof(uint16_t);

		/*    . */
		pData = (void*)printBinarySignals(pData, (uint16_t*)(pIDs + sizeof(uint16_t)),
			*(uint16_t*)pIDs, "Binary Outputs");
		pIDs += *(uint16_t*)pIDs * sizeof(uint16_t) + sizeof(uint16_t);

		/*     . */
		pData = (void*)printBinarySignals(pData, (uint16_t*)(pIDs + sizeof(uint16_t)),
			*(uint16_t*)pIDs, "Binary Input Signals");
		pIDs += *(uint16_t*)pIDs * sizeof(uint16_t) + sizeof(uint16_t);

		/*     . */
		pData = (void*)printBinarySignals(pData, (uint16_t*)(pIDs + sizeof(uint16_t)),
			*(uint16_t*)pIDs, "Binary Output Signals");
		pIDs += *(uint16_t*)pIDs * sizeof(uint16_t) + sizeof(uint16_t);

		/*     . */
		pData = (void*)printAnalogOutputSignals((const float*)pData, 
			(const mbf65_EventLogAOSDescriptorExp*)(pIDs + sizeof(uint16_t)),
			*(uint16_t*)pIDs, "Analog Output Signals");
		pIDs += *(uint16_t*)pIDs * sizeof(mbf65_EventLogAOSDescriptorExp) + sizeof(uint16_t);

		/*   . */
		pData = (void*)printAccumulators((const uint32_t*)pData, (uint16_t*)(pIDs + sizeof(uint16_t)),
			*(uint16_t*)pIDs, "Accumulators");
		pIDs += *(uint16_t*)pIDs * sizeof(uint16_t) + sizeof(uint16_t);

		pRecord++;
	}
} /* printEventLogRecords() */
/*-----------------------------------------------------------------------------------------------*/
const void* printBinarySignals(const void* pBitArray, const uint16_t* pIDArray, 
	uint16_t numberOfSignals, const char* pTitle)
{
	fprintf(stdout, "\n%s (%u):\n", pTitle, numberOfSignals);

	for (uint16_t signalIndex = 0; signalIndex < numberOfSignals; signalIndex++)
	{
		bool isBitSet = mbf65_getBinaryValueFromBitArray(pBitArray, signalIndex) != 0;
		if (isBitSet && pIDArray[signalIndex] != KIWI_INVALID_ID)
		{
			fprintf(stdout, "bitIndex=%u, ID=%u\n", signalIndex, pIDArray[signalIndex]);
		}
	}

	return (uint8_t*)pBitArray + numberOfSignals / 8 + 1;
} /* printBinarySignals() */
/*-----------------------------------------------------------------------------------------------*/
const void* printAnalogOutputSignals(const float* pValues, 
	const mbf65_EventLogAOSDescriptorExp* pDescriptors, uint16_t numberOfSignals, 
	const char* pTitle)
{
	fprintf(stdout, "\n%s (%u):\n", pTitle, numberOfSignals);

	for (uint16_t signalIndex = 0; signalIndex < numberOfSignals; signalIndex++)
	{
		float value = mbf65_readFloatFromBuf(pValues);
		if (!isnan(value) && pDescriptors[signalIndex].id != KIWI_INVALID_ID)
		{
			fprintf(stdout, "aosIndex=%u, aosID=%u, aosValue=%f, aosUnits=%u\n", signalIndex, 
				pDescriptors[signalIndex].id, value, pDescriptors[signalIndex].units);
		}

		pValues++;
	}

	return pValues;
} /* printAnalogOutputSignals() */
/*-----------------------------------------------------------------------------------------------*/
const void* printAccumulators(const uint32_t* pValues, const uint16_t* pIDArray,
	uint16_t numberOfSignals, const char* pTitle)
{
	fprintf(stdout, "\n%s (%u):\n", pTitle, numberOfSignals);

	for (uint16_t signalIndex = 0; signalIndex < numberOfSignals; signalIndex++)
	{
		uint32_t value = mbf65_readUint32FromBuf(pValues);
		if (value != 0xFFFFFFFFUL && pIDArray[signalIndex] != KIWI_INVALID_ID)
		{
			fprintf(stdout, "accumulatorIndex=%u, accumulatorID=%u, accumulatorValue=%u\n", 
				signalIndex, pIDArray[signalIndex], value);
		}

		pValues++;
	}

	return pValues;
} /* printAccumulators() */
/*-----------------------------------------------------------------------------------------------*/
